#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.commands`` -- Commands
"""

import sys

import edbob
from edbob import commands

import rattail


class Command(commands.Command):
    """
    The primary command for Rattail.
    """
    
    name = 'rattail'
    version = rattail.__version__
    description = "Retail Software Framework"
    long_description = """
Rattail is a retail software framework.

Copyright (c) 2010-2012 Lance Edgar <lance@edbob.org>

This program comes with ABSOLUTELY NO WARRANTY.  This is free software,
and you are welcome to redistribute it under certain conditions.
See the file COPYING.txt for more information.
"""


class DatabaseSyncCommand(commands.Subcommand):
    """
    Interacts with the database synchronization service; called as ``rattail
    dbsync``.
    """

    name = 'dbsync'
    description = "Manage the database synchronization service"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

    def run(self, args):
        from rattail.db.sync import linux as dbsync

        if args.subcommand == 'start':
            dbsync.start_daemon()

        elif args.subcommand == 'stop':
            dbsync.stop_daemon()


class FileMonitorCommand(commands.FileMonitorCommand):
    """
    Interacts with the file monitor service; called as ``rattail filemon``.

    See :class:`edbob.commands.FileMonitorCommand` for more information.
    """

    appname = 'rattail'

    def get_win32_module(self):
        from rattail import filemon
        return filemon

    def get_win32_service(self):
        from rattail.filemon import RattailFileMonitor
        return RattailFileMonitor


class LoadHostDataCommand(commands.Subcommand):
    """
    Loads data from the Rattail host database, if one is configured.
    """

    name = 'load-host-data'
    description = "Load data from host database"

    def run(self, args):
        from edbob.console import Progress
        from rattail.db import load

        edbob.init_modules(['edbob.db'])

        if 'host' not in edbob.engines:
            print "Host engine URL not configured."
            return

        proc = load.LoadProcessor()
        proc.load_all_data(edbob.engines['host'], Progress)


def main(*args):
    """
    The primary entry point for the Rattail command system.
    """

    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db.extension.model`` -- Schema Definition
"""

import logging

from sqlalchemy import Column, ForeignKey
from sqlalchemy import String, Integer, DateTime, Date, Boolean, Numeric, Text
from sqlalchemy import types
from sqlalchemy import and_
from sqlalchemy.orm import relationship, object_session
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.orderinglist import ordering_list

import edbob
from edbob.db.model import Base, uuid_column
from edbob.db.extensions.contact import Person, EmailAddress, PhoneNumber
from edbob.exceptions import LoadSpecError
from edbob.sqlalchemy import getset_factory

from rattail import sil
from rattail import batches
from rattail.gpc import GPCType


__all__ = ['Change', 'Store', 'StoreEmailAddress', 'StorePhoneNumber',
           'Department', 'Subdepartment', 'Brand', 'Category', 'Vendor',
           'VendorContact', 'VendorPhoneNumber', 'VendorEmailAddress',
           'Product', 'ProductCost', 'ProductPrice', 'Customer',
           'CustomerEmailAddress', 'CustomerPhoneNumber', 'CustomerGroup',
           'CustomerGroupAssignment', 'CustomerPerson', 'Employee',
           'EmployeeEmailAddress', 'EmployeePhoneNumber',
           'BatchColumn', 'Batch', 'BatchRow', 'LabelProfile']


log = logging.getLogger(__name__)


class Change(Base):
    """
    Represents a changed (or deleted) record, which is pending synchronization
    to another database.
    """

    __tablename__ = 'changes'

    class_name = Column(String(25), primary_key=True)
    uuid = Column(String(32), primary_key=True)
    deleted = Column(Boolean)

    def __repr__(self):
        status = 'deleted' if self.deleted else 'new/changed'
        return "<Change: %s, %s, %s>" % (self.class_name, self.uuid, status)


class BatchColumn(Base):
    """
    Represents a :class:`SilColumn` associated with a :class:`Batch`.
    """

    __tablename__ = 'batch_columns'

    uuid = uuid_column()
    batch_uuid = Column(String(32), ForeignKey('batches.uuid'))
    ordinal = Column(Integer, nullable=False)
    name = Column(String(20))
    display_name = Column(String(50))
    sil_name = Column(String(10))
    data_type = Column(String(15))
    description = Column(String(50))
    visible = Column(Boolean, default=True)

    def __init__(self, sil_name=None, **kwargs):
        if sil_name:
            kwargs['sil_name'] = sil_name
            sil_column = sil.get_column(sil_name)
            kwargs.setdefault('name', sil_name)
            kwargs.setdefault('data_type', sil_column.data_type)
            kwargs.setdefault('description', sil_column.description)
            kwargs.setdefault('display_name', sil_column.display_name)
        super(BatchColumn, self).__init__(**kwargs)

    def __repr__(self):
        return "<BatchColumn: %s>" % self.name

    def __unicode__(self):
        return unicode(self.display_name or '')


class BatchRow(edbob.Object):
    """
    Superclass of batch row objects.
    """

    def __unicode__(self):
        return u"Row %d" % self.ordinal


class Batch(Base):
    """
    Represents a SIL-compliant batch of data.
    """

    __tablename__ = 'batches'

    uuid = uuid_column()
    provider = Column(String(50))
    id = Column(String(8))
    source = Column(String(6))
    destination = Column(String(6))
    action_type = Column(String(6))
    description = Column(String(50))
    rowcount = Column(Integer, default=0)
    executed = Column(DateTime)
    purge = Column(Date)

    columns = relationship(BatchColumn, backref='batch',
                           collection_class=ordering_list('ordinal'),
                           order_by=BatchColumn.ordinal,
                           cascade='save-update, merge, delete, delete-orphan')

    _rowclasses = {}

    def __repr__(self):
        return "<Batch: %s>" % self.description

    def __unicode__(self):
        return unicode(self.description or '')

    @property
    def rowclass(self):
        """
        Returns the mapped class for the underlying row (data) table.
        """

        if not self.uuid:
            object_session(self).flush()
            assert self.uuid

        if self.uuid not in self._rowclasses:

            kwargs = {
                '__tablename__': 'batch.%s' % self.uuid,
                'uuid': uuid_column(),
                'ordinal': Column(Integer, nullable=False),
                }

            for column in self.columns:
                data_type = sil.get_sqlalchemy_type(column.data_type)
                kwargs[column.name] = Column(data_type)
            rowclass = type('BatchRow_%s' % str(self.uuid), (Base, BatchRow), kwargs)

            batch_uuid = self.uuid
            def batch(self):
                return object_session(self).query(Batch).get(batch_uuid)
            rowclass.batch = property(batch)

            self._rowclasses[self.uuid] = rowclass

        return self._rowclasses[self.uuid]

    def add_column(self, sil_name=None, **kwargs):
        column = BatchColumn(sil_name, **kwargs)
        self.columns.append(column)

    def add_row(self, row, **kwargs):
        """
        Adds a row to the batch data table.
        """

        session = object_session(self)
        # FIXME: This probably needs to use a func.max() query.
        row.ordinal = self.rowcount + 1
        session.add(row)
        self.rowcount += 1
        session.flush()

    def create_table(self):
        """
        Creates the batch's data table within the database.
        """

        session = object_session(self)
        self.rowclass.__table__.create(session.bind)

    def drop_table(self):
        """
        Drops the batch's data table from the database.
        """

        log.debug("Batch.drop_table: Dropping table for batch: %s, %s (%s)"
                  % (self.id, self.description, self.uuid))
        session = object_session(self)
        self.rowclass.__table__.drop(session.bind)

    def execute(self, progress=None):
        provider = self.get_provider()
        assert provider
        if not provider.execute(self, progress):
            return False
        self.executed = edbob.utc_time(naive=True)
        object_session(self).flush()
        return True

    def get_provider(self):
        assert self.provider
        return batches.get_provider(self.provider)

    def iter_rows(self):
        session = object_session(self)
        q = session.query(self.rowclass)
        q = q.order_by(self.rowclass.ordinal)
        return q


class StoreEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Store`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Store'}


class StorePhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Store`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Store'}


class Store(Base):
    """
    Represents a store (physical or otherwise) within the organization.
    """

    __tablename__ = 'stores'

    uuid = uuid_column()
    id = Column(String(10))
    name = Column(String(100))

    def __repr__(self):
        return "<Store: %s, %s>" % (self.id, self.name)

    def __unicode__(self):
        return unicode(self.name or '')

    def add_email_address(self, address, type='Info'):
        email = StoreEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Voice'):
        phone = StorePhoneNumber(number=number, type=type)
        self.phones.append(phone)

Store.emails = relationship(
    StoreEmailAddress,
    backref='store',
    primaryjoin=StoreEmailAddress.parent_uuid == Store.uuid,
    foreign_keys=[StoreEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=StoreEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Store.email = relationship(
    StoreEmailAddress,
    primaryjoin=and_(
        StoreEmailAddress.parent_uuid == Store.uuid,
        StoreEmailAddress.preference == 1),
    foreign_keys=[StoreEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)

Store.phones = relationship(
    StorePhoneNumber,
    backref='store',
    primaryjoin=StorePhoneNumber.parent_uuid == Store.uuid,
    foreign_keys=[StorePhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=StorePhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Store.phone = relationship(
    StorePhoneNumber,
    primaryjoin=and_(
        StorePhoneNumber.parent_uuid == Store.uuid,
        StorePhoneNumber.preference == 1),
    foreign_keys=[StorePhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class Brand(Base):
    """
    Represents a brand or similar product line.
    """

    __tablename__ = 'brands'

    uuid = uuid_column()
    name = Column(String(100))

    def __repr__(self):
        return "<Brand: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')


class Department(Base):
    """
    Represents an organizational department.
    """

    __tablename__ = 'departments'

    uuid = uuid_column()
    number = Column(Integer)
    name = Column(String(30))

    def __repr__(self):
        return "<Department: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')


class Subdepartment(Base):
    """
    Represents an organizational subdepartment.
    """

    __tablename__ = 'subdepartments'

    uuid = uuid_column()
    number = Column(Integer)
    name = Column(String(30))
    department_uuid = Column(String(32), ForeignKey('departments.uuid'))

    department = relationship(Department)

    def __repr__(self):
        return "<Subdepartment: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')


class Category(Base):
    """
    Represents an organizational category for products.
    """

    __tablename__ = 'categories'

    uuid = uuid_column()
    number = Column(Integer)
    name = Column(String(50))
    department_uuid = Column(String(32), ForeignKey('departments.uuid'))

    department = relationship(Department)

    def __repr__(self):
        return "<Category: %s, %s>" % (self.number, self.name)

    def __str__(self):
        return str(self.name or '')

    def __unicode__(self):
        return unicode(self.name or '')


class VendorEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Vendor`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Vendor'}


class VendorPhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Vendor`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Vendor'}


class VendorContact(Base):
    """
    Represents a point of contact (e.g. salesperson) for a vendor, from the
    retailer's perspective.
    """

    __tablename__ = 'vendor_contacts'

    uuid = uuid_column()
    vendor_uuid = Column(String(32), ForeignKey('vendors.uuid'), nullable=False)
    person_uuid = Column(String(32), ForeignKey('people.uuid'), nullable=False)
    preference = Column(Integer, nullable=False)

    person = relationship(Person)

    def __repr__(self):
        return "<VendorContact: %s, %s>" % (self.vendor, self.person)

    def __unicode__(self):
        return unicode(self.person)


class Vendor(Base):
    """
    Represents a vendor from which products are purchased by the store.
    """

    __tablename__ = 'vendors'

    uuid = uuid_column()
    id = Column(String(15))
    name = Column(String(40))
    special_discount = Column(Numeric(5,3))

    contacts = relationship(VendorContact, backref='vendor',
                            collection_class=ordering_list('preference', count_from=1),
                            order_by=VendorContact.preference,
                            cascade='save-update, merge, delete, delete-orphan')

    def __repr__(self):
        return "<Vendor: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')

    def add_contact(self, person):
        contact = VendorContact(person=person)
        self.contacts.append(contact)

    def add_email_address(self, address, type='Info'):
        email = VendorEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Voice'):
        phone = VendorPhoneNumber(number=number, type=type)
        self.phones.append(phone)

Vendor.emails = relationship(
    VendorEmailAddress,
    backref='vendor',
    primaryjoin=VendorEmailAddress.parent_uuid == Vendor.uuid,
    foreign_keys=[VendorEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=VendorEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Vendor.email = relationship(
    VendorEmailAddress,
    primaryjoin=and_(
        VendorEmailAddress.parent_uuid == Vendor.uuid,
        VendorEmailAddress.preference == 1),
    foreign_keys=[VendorEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)

Vendor.phones = relationship(
    VendorPhoneNumber,
    backref='vendor',
    primaryjoin=VendorPhoneNumber.parent_uuid == Vendor.uuid,
    foreign_keys=[VendorPhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=VendorPhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Vendor.phone = relationship(
    VendorPhoneNumber,
    primaryjoin=and_(
        VendorPhoneNumber.parent_uuid == Vendor.uuid,
        VendorPhoneNumber.preference == 1),
    foreign_keys=[VendorPhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)

Vendor.contact = relationship(
    VendorContact,
    primaryjoin=and_(
        VendorContact.vendor_uuid == Vendor.uuid,
        VendorContact.preference == 1),
    uselist=False,
    viewonly=True)


class ProductCost(Base):
    """
    Represents a source from which a product may be obtained via purchase.
    """

    __tablename__ = 'product_costs'

    uuid = uuid_column()
    product_uuid = Column(String(32), ForeignKey('products.uuid'), nullable=False)
    vendor_uuid = Column(String(32), ForeignKey('vendors.uuid'), nullable=False)
    preference = Column(Integer, nullable=False)
    code = Column(String(15))

    case_size = Column(Integer)
    case_cost = Column(Numeric(9,5))
    pack_size = Column(Integer)
    pack_cost = Column(Numeric(9,5))
    unit_cost = Column(Numeric(9,5))
    effective = Column(DateTime)

    vendor = relationship(Vendor)

    def __repr__(self):
        return "<ProductCost: %s : %s>" % (self.product, self.vendor)


class ProductPrice(Base):
    """
    Represents a price for a product.
    """

    __tablename__ = 'product_prices'

    uuid = uuid_column()
    product_uuid = Column(String(32), ForeignKey('products.uuid'), nullable=False)
    type = Column(Integer)
    level = Column(Integer)
    starts = Column(DateTime)
    ends = Column(DateTime)
    price = Column(Numeric(8,3))
    multiple = Column(Integer)
    pack_price = Column(Numeric(8,3))
    pack_multiple = Column(Integer)

    def __repr__(self):
        return "<ProductPrice: %s : %s>" % (self.product, self.price)


class Product(Base):
    """
    Represents a product for sale and/or purchase.
    """

    __tablename__ = 'products'

    uuid = uuid_column()
    upc = Column(GPCType, index=True)
    department_uuid = Column(String(32), ForeignKey('departments.uuid'))
    subdepartment_uuid = Column(String(32), ForeignKey('subdepartments.uuid'))
    category_uuid = Column(String(32), ForeignKey('categories.uuid'))
    brand_uuid = Column(String(32), ForeignKey('brands.uuid'))
    description = Column(String(60))
    description2 = Column(String(60))
    size = Column(String(30))
    unit_of_measure = Column(String(4))

    regular_price_uuid = Column(
        String(32),
        ForeignKey('product_prices.uuid',
                   use_alter=True,
                   name='products_regular_price_uuid_fkey'))

    current_price_uuid = Column(
        String(32),
        ForeignKey('product_prices.uuid',
                   use_alter=True,
                   name='products_current_price_uuid_fkey'))

    department = relationship(Department)
    subdepartment = relationship(Subdepartment)
    category = relationship(Category)
    brand = relationship(Brand)

    costs = relationship(ProductCost, backref='product',
                         collection_class=ordering_list('preference', count_from=1),
                         order_by=ProductCost.preference,
                         cascade='save-update, merge, delete, delete-orphan')

    def __repr__(self):
        return "<Product: %s>" % self.description

    def __unicode__(self):
        return unicode(self.description or '')

Product.prices = relationship(
    ProductPrice, backref='product',
    primaryjoin=ProductPrice.product_uuid == Product.uuid,
    cascade='save-update, merge, delete, delete-orphan')

Product.regular_price = relationship(
    ProductPrice,
    primaryjoin=Product.regular_price_uuid == ProductPrice.uuid,
    lazy='joined',
    post_update=True)

Product.current_price = relationship(
    ProductPrice,
    primaryjoin=Product.current_price_uuid == ProductPrice.uuid,
    lazy='joined',
    post_update=True)

Product.cost = relationship(
    ProductCost,
    primaryjoin=and_(
        ProductCost.product_uuid == Product.uuid,
        ProductCost.preference == 1,
        ),
    uselist=False,
    viewonly=True)

Product.vendor = relationship(
    Vendor,
    secondary=ProductCost.__table__,
    primaryjoin=and_(
        ProductCost.product_uuid == Product.uuid,
        ProductCost.preference == 1,
        ),
    secondaryjoin=Vendor.uuid == ProductCost.vendor_uuid,
    uselist=False,
    viewonly=True)


class EmployeeEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Employee`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Employee'}


class EmployeePhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Employee`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Employee'}


class Employee(Base):
    """
    Represents an employee within the organization.
    """

    __tablename__ = 'employees'

    uuid = uuid_column()
    id = Column(Integer)
    person_uuid = Column(String(32), ForeignKey('people.uuid'), nullable=False)
    status = Column(Integer)
    display_name = Column(String(100))

    person = relationship(Person)

    first_name = association_proxy('person', 'first_name')
    last_name = association_proxy('person', 'last_name')

    def __repr__(self):
        return "<Employee: %s>" % self.person

    def __unicode__(self):
        return unicode(self.display_name or self.person)

    def add_email_address(self, address, type='Home'):
        email = EmployeeEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Home'):
        phone = EmployeePhoneNumber(number=number, type=type)
        self.phones.append(phone)

Employee.emails = relationship(
    EmployeeEmailAddress,
    backref='employee',
    primaryjoin=EmployeeEmailAddress.parent_uuid == Employee.uuid,
    foreign_keys=[EmployeeEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=EmployeeEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Employee.email = relationship(
    EmployeeEmailAddress,
    primaryjoin=and_(
        EmployeeEmailAddress.parent_uuid == Employee.uuid,
        EmployeeEmailAddress.preference == 1),
    foreign_keys=[EmployeeEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)

Employee.phones = relationship(
    EmployeePhoneNumber,
    backref='employee',
    primaryjoin=EmployeePhoneNumber.parent_uuid == Employee.uuid,
    foreign_keys=[EmployeePhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=EmployeePhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Employee.phone = relationship(
    EmployeePhoneNumber,
    primaryjoin=and_(
        EmployeePhoneNumber.parent_uuid == Employee.uuid,
        EmployeePhoneNumber.preference == 1),
    foreign_keys=[EmployeePhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class CustomerEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Customer`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Customer'}


class CustomerPhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Customer`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Customer'}


class CustomerPerson(Base):
    """
    Represents the association between a :class:`Person` and a customer account
    (:class:`Customer`).
    """

    __tablename__ = 'customers_people'

    uuid = uuid_column()
    customer_uuid = Column(String(32), ForeignKey('customers.uuid'))
    person_uuid = Column(String(32), ForeignKey('people.uuid'))
    ordinal = Column(Integer, nullable=False)

    person = relationship(Person)


class Customer(Base):
    """
    Represents a customer account.  Customer accounts may consist of more than
    one :class:`Person`, in some cases.
    """

    __tablename__ = 'customers'

    uuid = uuid_column()
    id = Column(String(20))
    name = Column(String(255))
    email_preference = Column(Integer)

    def __repr__(self):
        return "<Customer: %s, %s>" % (self.id, self.name or self.person)

    def __unicode__(self):
        return unicode(self.name or self.person)

    def add_email_address(self, address, type='Home'):
        email = CustomerEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Home'):
        phone = CustomerPhoneNumber(number=number, type=type)
        self.phones.append(phone)

Customer.emails = relationship(
    CustomerEmailAddress,
    backref='customer',
    primaryjoin=CustomerEmailAddress.parent_uuid == Customer.uuid,
    foreign_keys=[CustomerEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=CustomerEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Customer.email = relationship(
    CustomerEmailAddress,
    primaryjoin=and_(
        CustomerEmailAddress.parent_uuid == Customer.uuid,
        CustomerEmailAddress.preference == 1),
    foreign_keys=[CustomerEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)

Customer.phones = relationship(
    CustomerPhoneNumber,
    backref='customer',
    primaryjoin=CustomerPhoneNumber.parent_uuid == Customer.uuid,
    foreign_keys=[CustomerPhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=CustomerPhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Customer.phone = relationship(
    CustomerPhoneNumber,
    primaryjoin=and_(
        CustomerPhoneNumber.parent_uuid == Customer.uuid,
        CustomerPhoneNumber.preference == 1),
    foreign_keys=[CustomerPhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)

Customer._people = relationship(
    CustomerPerson, backref='customer',
    primaryjoin=CustomerPerson.customer_uuid == Customer.uuid,
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=CustomerPerson.ordinal)

Customer.people = association_proxy(
    '_people', 'person',
    getset_factory=getset_factory,
    creator=lambda x: CustomerPerson(person=x))

Customer._person = relationship(
    CustomerPerson,
    primaryjoin=and_(
        CustomerPerson.customer_uuid == Customer.uuid,
        CustomerPerson.ordinal == 1),
    uselist=False,
    viewonly=True)

Customer.person = association_proxy(
    '_person', 'person',
    getset_factory=getset_factory)


class CustomerGroup(Base):
    """
    Represents an arbitrary group to which :class:`Customer` instances may (or
    may not) belong.
    """

    __tablename__ = 'customer_groups'

    uuid = uuid_column()
    id = Column(String(20))
    name = Column(String(255))

    def __repr__(self):
        return "<CustomerGroup: %s, %s>" % (self.id, self.name)

    def __unicode__(self):
        return unicode(self.name or '')


class CustomerGroupAssignment(Base):
    """
    Represents the assignment of a :class:`Customer` to a
    :class:`CustomerGroup`.
    """

    __tablename__ = 'customers_groups'

    uuid = uuid_column()
    customer_uuid = Column(String(32), ForeignKey('customers.uuid'))
    group_uuid = Column(String(32), ForeignKey('customer_groups.uuid'))
    ordinal = Column(Integer, nullable=False)

    group = relationship(CustomerGroup)

Customer._groups = relationship(
    CustomerGroupAssignment, backref='customer',
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=CustomerGroupAssignment.ordinal)

Customer.groups = association_proxy(
    '_groups', 'group',
    getset_factory=getset_factory,
    creator=lambda x: CustomerGroupAssignment(group=x))


class LabelProfile(Base):
    """
    Represents a "profile" (collection of settings) for product label printing.
    """

    __tablename__ = 'label_profiles'

    uuid = uuid_column()
    ordinal = Column(Integer)
    code = Column(String(3))
    description = Column(String(50))
    printer_spec = Column(String(255))
    formatter_spec = Column(String(255))
    format = Column(Text)

    _printer = None
    _formatter = None

    def __repr__(self):
        return "<LabelProfile: %s>" % self.code

    def __unicode__(self):
        return unicode(self.description or '')

    def get_formatter(self):
        if not self._formatter and self.formatter_spec:
            try:
                formatter = edbob.load_spec(self.formatter_spec)
            except LoadSpecError:
                pass
            else:
                self._formatter = formatter()
        return self._formatter        

    def get_printer(self):
        if not self._printer and self.printer_spec:
            try:
                printer = edbob.load_spec(self.printer_spec)
            except LoadSpecError:
                pass
            else:
                self._printer = printer()
                for name in printer.required_settings:
                    setattr(printer, name, self.get_printer_setting(name))
                self._printer.formatter = self.get_formatter()
        return self._printer

    def get_printer_setting(self, name):
        session = object_session(self)
        if not self.uuid:
            session.flush()
        name = 'labels.%s.printer.%s' % (self.uuid, name)
        return edbob.get_setting(name, session)

    def save_printer_setting(self, name, value):
        session = object_session(self)
        if not self.uuid:
            session.flush()
        name = 'labels.%s.printer.%s' % (self.uuid, name)
        edbob.save_setting(name, value, session)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.batches.providers`` -- Batch Providers
"""

import datetime

import edbob

import rattail
from rattail import sil


__all__ = ['BatchProvider']


class BatchProvider(edbob.Object):

    name = None
    description = None
    source = 'RATAIL'
    destination = None
    action_type = None
    purge_date_offset = 90

    session = None

    def add_columns(self, batch):
        pass

    def add_rows_begin(self, batch, data):
        pass

    def add_rows(self, batch, data, progress=None):

        result = self.add_rows_begin(batch, data)
        if result is not None and not result:
            return False

        prog = None
        if progress:
            prog = progress("Adding rows to batch \"%s\"" % batch.description,
                            data.count())
        cancel = False
        for i, datum in enumerate(data, 1):
            self.add_row(batch, datum)
            if prog and not prog.update(i):
                cancel = True
                break
        if prog:
            prog.destroy()

        if not cancel:
            result = self.add_rows_end(batch, progress)
            if result is not None:
                cancel = not result

        return not cancel

    def add_rows_end(self, batch, progress=None):
        pass

    def execute(self, batch, progress=None):
        raise NotImplementedError

    def make_batch(self, session, data, progress=None):
        self.session = session

        batch = rattail.Batch()
        batch.provider = self.name
        batch.source = self.source
        batch.id = sil.consume_batch_id(batch.source)
        batch.destination = self.destination
        batch.description = self.description
        batch.action_type = self.action_type
        self.set_purge_date(batch)
        session.add(batch)
        session.flush()

        self.add_columns(batch)
        batch.create_table()
        if not self.add_rows(batch, data, progress=progress):
            batch.drop_table()
            return None
        return batch

    def set_purge_date(self, batch):
        today = edbob.utc_time(naive=True).date()
        purge_offset = datetime.timedelta(days=self.purge_date_offset)
        batch.purge = today + purge_offset

    def set_params(self, session, **params):
        pass



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db.extension.enum`` -- Enumerations
"""


EMPLOYEE_STATUS_CURRENT         = 1
EMPLOYEE_STATUS_FORMER          = 2

EMPLOYEE_STATUS = {
    EMPLOYEE_STATUS_CURRENT     : "current",
    EMPLOYEE_STATUS_FORMER      : "former",
    }


# VENDOR_CATALOG_NOT_PARSED       = 1
# VENDOR_CATALOG_PARSED           = 2
# VENDOR_CATALOG_COGNIZED         = 3
# VENDOR_CATALOG_PROCESSED        = 4

# VENDOR_CATALOG_STATUS = {
#     VENDOR_CATALOG_NOT_PARSED   : "not parsed",
#     VENDOR_CATALOG_PARSED       : "parsed",
#     VENDOR_CATALOG_COGNIZED     : "cognized",
#     VENDOR_CATALOG_PROCESSED    : "processed",
#     }



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.db.sync`` -- Database Synchronization
"""

import sys
import time
import logging

if sys.platform == 'win32':
    import win32api

from sqlalchemy.orm import class_mapper

import edbob

import rattail


log = logging.getLogger(__name__)


def get_sync_engines():
    edbob.init_modules(['edbob.db'])

    keys = edbob.config.get('rattail.db', 'syncs')
    if not keys:
        return None

    engines = {}
    for key in keys.split(','):
        key = key.strip()
        engines[key] = edbob.engines[key]
    log.info("get_sync_engines: Found engine keys: %s" % ','.join(engines.keys()))
    return engines


def dependency_sort(x, y):
    map_x = class_mapper(getattr(edbob, x))
    map_y = class_mapper(getattr(edbob, y))

    dep_x = []
    table_y = map_y.tables[0].name
    for column in map_x.columns:
        for key in column.foreign_keys:
            if key.column.table.name == table_y:
                return 1
            dep_x.append(key)

    dep_y = []
    table_x = map_x.tables[0].name
    for column in map_y.columns:
        for key in column.foreign_keys:
            if key.column.table.name == table_x:
                return -1
            dep_y.append(key)

    if dep_x and not dep_y:
        return 1
    if dep_y and not dep_x:
        return -1
    return 0


def delete_instance(session, instance):
    """
    Delete ``instance`` using ``session``.

    This function is somewhat of a hack, in order to properly handle certain
    foreign key dependencies in the Rattail data model.
    """

    if isinstance(instance, rattail.Department):

        # Remove association from Subdepartment records.
        q = session.query(rattail.Subdepartment)
        q = q.filter(rattail.Subdepartment.department == instance)
        for subdepartment in q:
            subdepartment.department = None
            
        # Remove association from Product records.
        q = session.query(rattail.Product)
        q = q.filter(rattail.Product.department == instance)
        for product in q:
            product.department = None

    elif isinstance(instance, rattail.Subdepartment):

        # Remove association from Product records.
        q = session.query(rattail.Product)
        q = q.filter(rattail.Product.subdepartment == instance)
        for product in q:
            product.subdepartment = None

    elif isinstance(instance, rattail.Vendor):

        # Remove associated ProductCost records.
        q = session.query(rattail.ProductCost)
        q = q.filter(rattail.ProductCost.vendor == instance)
        for cost in q:
            session.delete(cost)

    session.delete(instance)


def merge_instance(session, instance):
    """
    Merge ``instance`` into ``session``.

    This function is somewhat of a hack, in order to properly handle
    :class:`rattail.Product` instances and the interdependent nature of the
    related :class:`rattail.ProductPrice` instances.
    """

    if not isinstance(instance, rattail.Product):
        return session.merge(instance)

    product = session.merge(instance)

    # I'm not 100% sure I understand this correctly, but here's my thinking:
    # First we clear the price relationships in case they've actually gone
    # away; then we reestablish any which are currently valid.

    # Setting the price relationship attributes to ``None`` isn't enough to
    # force the ORM to notice a change, since the uuid field is ultimately what
    # it's watching.  So we must explicitly use that field here.
    product.regular_price_uuid = None
    product.current_price_uuid = None

    # If the source instance has currently valid price relationships, then we
    # reestablish them.  We must merge the source price instance in order to be
    # certain it will exist in the target session, and avoid foreign key
    # errors.  However we *still* must also set the uuid fields because again,
    # the ORM is watching those...  This was noticed to be the source of some
    # bugs where successive database syncs were effectively "toggling" the
    # price relationship.  Setting the uuid field explicitly seems to solve it.
    if instance.regular_price_uuid:
        product.regular_price = session.merge(instance.regular_price)
        product.regular_price_uuid = product.regular_price.uuid
    if instance.current_price_uuid:
        product.current_price = session.merge(instance.current_price)
        product.current_price_uuid = product.current_price.uuid

    return product


def synchronize_changes(engines):

    log.info("synchronize_changes: Using engine keys: %s" % ','.join(engines.keys()))

    while True:
        local_session = edbob.Session()
        local_changes = local_session.query(rattail.Change)

        if local_changes.count():
            log.debug("synchronize_changes: found %d changes" % local_changes.count())

            # First we must determine which types of instances are in need of
            # syncing.  The order will matter because of class dependencies.
            # However the dependency_sort() call doesn't *quite* take care of
            # everything - notably the Product/ProductPrice situation.  Since
            # those classes are mutually dependent, we start with a hackish
            # lexical sort and hope for the best...
            q = local_session.query(rattail.Change.class_name.distinct())
            q = q.order_by('class_name')
            class_names = [x[0] for x in q]
            class_names.sort(cmp=dependency_sort)

            remote_sessions = []
            for remote_engine in engines.itervalues():
                remote_sessions.append(
                    edbob.Session(bind=remote_engine))

            for class_name in class_names:

                for change in local_changes.filter_by(class_name=class_name):
                    log.debug("synchronize_changes: processing change: %s" % change)
                    cls = getattr(edbob, change.class_name)

                    if change.deleted:
                        for remote_session in remote_sessions:
                            remote_instance = remote_session.query(cls).get(change.uuid)
                            if remote_instance:
                                delete_instance(remote_session, remote_instance)
                                remote_session.flush()

                    else: # new/dirty
                        local_instance = local_session.query(cls).get(change.uuid)
                        if local_instance:
                            for remote_session in remote_sessions:
                                merge_instance(remote_session, local_instance)
                                remote_session.flush()

                    local_session.delete(change)
                    local_session.flush()

            for remote_session in remote_sessions:
                remote_session.commit()
                remote_session.close()
            local_session.commit()

        local_session.close()
        if sys.platform == 'win32':
            win32api.Sleep(3000)
        else:
            time.sleep(3)
